/*
 *
 * A simple program that removes a file in a deterministic location.
 *
 * It should be SETUID root or have the right CAPABILITIES(7).
 *
*/

/*

Copyright 2017 Fermi Research Alliance, LLC

This software was produced under U.S. Government contract DE-AC02-07CH11359 for Fermi National Accelerator Laboratory (Fermilab), which is operated by Fermi Research Alliance, LLC for the U.S. Department of Energy. The U.S. Government has rights to use, reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR FERMI RESEARCH ALLIANCE, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is modified to produce derivative works, such modified software should be clearly marked, so as not to confuse it with the version available from Fermilab.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR FERMI RESEARCH ALLIANCE, LLC BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#ifndef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 200112L
#endif

#ifndef _GNU_SOURCE
#define _GNU_SOURCE 0
#endif

#include <pwd.h>        /* for getpwuid, passwd                 */
#include <stdio.h>      /* for fprintf, fwrite, stderr, etc     */
#include <string.h>     /* for basename, strncat, memset        */
#include <sys/stat.h>   /* for stat, chmod, S_IRUSR, etc        */
#include <sys/prctl.h>  /* for prctl, PR_SET_DUMPABLE           */
#include <sys/types.h>  /* for uid_t, cap_t, etc                */
#include <unistd.h>     /* for gethostname, getuid, etc         */


#ifdef CAPABILITIES
 #include <sys/capability.h>
 cap_value_t expected_capabilities[] = {CAP_DAC_OVERRIDE, CAP_FOWNER};
 int num_caps = sizeof(expected_capabilities) / sizeof(expected_capabilities[0]);

 static int disable_capabilities(void) {
   cap_t capabilities;

   capabilities = cap_get_proc();

   if (cap_clear(capabilities)) {
     cap_free(capabilities);
     /* error */
     fprintf(stderr, "remove-kcron-keytab: Unable to clear CAPABILITIES\n");
     return 1;
   }

   cap_free(capabilities);
   return 0;
 }

 static int enable_capabilities(void) {
   cap_t capabilities;
   capabilities = cap_get_proc();

   /* clear any active capabilities */
   int clear_cap = disable_capabilities();
   if (clear_cap != 0) {
     cap_free(capabilities);
     return clear_cap;
   }

   if (cap_set_flag(capabilities, CAP_PERMITTED, num_caps, expected_capabilities, CAP_SET) == -1) {
     cap_free(capabilities);
     /* error */
     fprintf(stderr, "remove-kcron-keytab: Unable to set CAPABILITIES PERMITTED\n");
     return 1;
   }
   if (cap_set_flag(capabilities, CAP_EFFECTIVE, num_caps, expected_capabilities, CAP_SET) == -1) {
     cap_free(capabilities);
     /* error */
     fprintf(stderr, "remove-kcron-keytab: Unable to set CAPABILITIES EFFECTIVE\n");
     return 1;
   }
   if (cap_set_proc(capabilities) == -1) {
     cap_free(capabilities);
     /* error */
     fprintf(stderr, "remove-kcron-keytab: Unable to activate CAPABILITIES\n");
     return 1;
   }

   cap_free(capabilities);
   return 0;
 }

#endif

int main (void) {

  static int result = 0;

#ifdef CAPABILITIES
  int setcap_result = 0;
#endif

  if (0 > prctl(PR_SET_DUMPABLE, 0)) {
    fprintf(stderr, "remove-kcron-keytab: Cannot disable core dumps.\n");
    return 1;
  }

  /* What is this system called? */
  char hostname[300];
  memset(hostname, '\0', sizeof(hostname));
  if (0 != (gethostname(hostname, 298))) {
    fprintf(stderr, "remove-kcron-keytab: gethostname() error.\n");
    return 1;
  }


  /* What is my UID (not effective UID), ie whoami when I'm not root */
  uid_t uid;
  uid = getuid ();
  struct passwd *pd;
  if (NULL == (pd = getpwuid(uid))) {
    fprintf(stderr, "remove-kcron-keytab: getpwuid() error for %d.\n", uid);
    return 1;
  }
  char username[300];
  memset(username, '\0', sizeof(username));
  strncat(username, pd->pw_name, 298);

  /* Where do the keytabs go?  Here of course */
  char keytab[1000];
  memset(keytab, '\0', sizeof(keytab));
  strncat(keytab, "/var/adm/krb5/", 60);

  /* all done keytab dir if missing */
  struct stat st = {0};
  if (stat(keytab, &st) == -1) {
    return 0;
  }

  /* Add in the filename */
  strncat(keytab, basename(username), 300);
  strncat(keytab, ".cron.", 7);
  strncat(keytab, basename(hostname), 300);
  strncat(keytab, ".keytab", 8);

  /* If keytab is missing we are done */
  if (stat(keytab, &st) == -1) {
    return 0;
  } else {
#ifdef CAPABILITIES
    setcap_result = enable_capabilities ();
    if (setcap_result != 0 ) {
      return 1;
    }
#endif
    result = remove(keytab);
#ifdef CAPABILITIES
    setcap_result = disable_capabilities ();
    if (setcap_result != 0 ) {
      return 1;
    }
#endif
    if (result != 0) {
      fprintf(stderr, "remove-kcron-keytab: Failed: rm %s\n", keytab);
      return 1;
    }
  }

  return 0;
}
